/**
 * $Id$
 *
 * Copyright (C) shinichiro.h <hamaji@nii.ac.jp>
 *  http://shinh.skr.jp/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdarg.h>

#include <tidy.h>
#include <buffio.h>

#include "gui.h"
#include "rc.h"

#include "objalloc.h"
#include "hashtab.h"

typedef struct objalloc objalloc;

typedef enum {
    TEXT, CHECKBOX, RADIO, SELECT
} InputType;

typedef struct {
    InputType type;
    widget* widget;
    char* name;
    int value_size;
    union {
        char* value;
        char** values;
    };
} Input;

htab_t inputs;

RC* rc;

int strcasecmp(const char* s1, const char* s2);

void die(const char* format, ...) {
    va_list ap;
    va_start(ap, format);
    vfprintf(stderr, format, ap);
    va_end(ap);
    exit(1);
}

char* objstrcpy(objalloc* pool, const char* src) {
    char* dst = (char*)objalloc_alloc(pool, sizeof(char)*(strlen(src)+1));
    strcpy(dst, src);
    return dst;
}

void chomp(char* str) {
    int i;
    if (!str || !*str) return;
    i = strlen(str)-1;
    if (str[i] == '\n') {
        str[i] = '\0';
        i--;
    }
    if (!*str) return;
    if (str[i] == '\r') {
        str[i] = '\0';
        i--;
    }
}

char* tidy_get_text(objalloc* pool, TidyDoc tdoc, TidyNode node) {
    TidyBuffer buf;
    char* text;
    tidyBufInit(&buf);
    tidyNodeGetText(tdoc, node, &buf);
    text = objstrcpy(pool, (char*)buf.bp);
    chomp(text);
    tidyBufFree(&buf);
    return text;
}

char* tidy_get_title(objalloc* pool, TidyDoc tdoc) {
    TidyNode node;
    node = tidyGetHead(tdoc);
    for (node = tidyGetChild(node); node; node = tidyGetNext(node)) {
        if (tidyNodeIsTITLE(node)) {
            return tidy_get_text(pool, tdoc, tidyGetChild(node));
        }
    }
    return NULL;
}

static TidyNode tidy_node_next(TidyNode node) {
    TidyNode next;
    next = tidyGetNext(node);
    while (!next) {
        node = tidyGetParent(node);
        if (tidyNodeIsBODY(node)) return NULL;
        next = tidyGetNext(node);
    }
    return next;
}

static TidyNode tidy_node_deep(TidyNode node) {
    TidyNode next;
    next = tidyGetChild(node);
    if (!next) next = tidy_node_next(node);
    return next;
}

TidyNode tidy_search_node(TidyNode node, Bool (*pred) (TidyNode)) {
    for (; node; node = tidy_node_deep(node)) {
        if (pred(node)) return node;
    }
    return NULL;
}

widget* widgets_get(objalloc* pool, TidyDoc tdoc, TidyNode form);

widget* widget_get(objalloc* pool, TidyDoc tdoc, TidyNode node) {
    if (tidyNodeIsText(node)) {
        char* text = tidy_get_text(pool, tdoc, node);
        return gui_label_new(text);
    }
    else if (tidyNodeIsINPUT(node)) {
        char* type = (char*)tidyAttrValue(tidyAttrGetTYPE(node));
        char* name = (char*)tidyAttrValue(tidyAttrGetNAME(node));
        char* value = (char*)tidyAttrValue(tidyAttrGetVALUE(node));
        if (!value) value = "1";

        Input** ent;
        InputType input_type = TEXT;
        widget* w = NULL;

        const char* rc_value;

        if (type && strcasecmp(type, "submit") == 0) {
            return gui_button_new(value);
        }

        if (!name) {
            die("input (type=%s) should have name attr\n", type);
        }
        ent = (Input**)htab_find_slot(inputs, name, INSERT);

        rc_value = RC_Get(rc, name);

        if (type == NULL || strcasecmp(type, "text") == 0) {
            if (*ent) die("input (name=%s) comes twice\n", name);
            input_type = TEXT;
            w = gui_text_new(rc_value);
        }
        else if (strcasecmp(type, "checkbox") == 0) {
            if (*ent) die("input (name=%s) comes twice\n", name);
            input_type = CHECKBOX;
            w = gui_checkbox_new(rc_value && strcmp(rc_value, value) == 0);
        }
        else if (strcasecmp(type, "radio") == 0) {
            w = gui_radio_new((*ent) ? (*ent)->widget : NULL,
                              rc_value && strcmp(rc_value, value) == 0);

            if (!*ent) {
                *ent = (Input*)objalloc_alloc(pool, sizeof(Input));
                (*ent)->type = RADIO;
                (*ent)->name = objstrcpy(pool, name);
                (*ent)->widget = w;
                (*ent)->value_size = 0;
                (*ent)->values = NULL;
            }
            (*ent)->value_size++;
            (*ent)->values = (char**)realloc((*ent)->values,
                                             sizeof(char*)*(*ent)->value_size);
            (*ent)->values[(*ent)->value_size-1] = objstrcpy(pool, value);

            return w;
        }
        else {
            die("input (type=%s) is unknown\n", type);
        }

        *ent = (Input*)objalloc_alloc(pool, sizeof(Input));
        (*ent)->type = input_type;
        (*ent)->name = objstrcpy(pool, name);
        (*ent)->widget = w;

        if (input_type == CHECKBOX) {
            (*ent)->value = objstrcpy(pool, value);
        }

        return w;
    }
    else if (tidyNodeIsSELECT(node)) {
        char* name = (char*)tidyAttrValue(tidyAttrGetNAME(node));
        TidyNode opt;
        char** list = NULL;
        int size = 0;
        widget* select;
        Input** ent;
        int selected = 0;
        const char* rc_value;
        char** values = NULL;

        if (!name) {
            die("select should have name attr\n");
        }

        rc_value = RC_Get(rc, name);

        for (opt = tidyGetChild(node); opt; opt = tidyGetNext(opt)) {
            char* text = tidy_get_text(pool, tdoc, tidyGetChild(opt));
            char* opt_val = (char*)tidyAttrValue(tidyAttrGetVALUE(opt));

            if (!opt_val) {
                die("option should have value attr\n");
            }
            if (rc_value && strcmp(opt_val, rc_value) == 0) {
                selected = size;
            }

            size++;
            list = (char**)realloc(list, sizeof(char*)*size);
            list[size-1] = text;
            values = (char**)realloc(values, sizeof(char*)*size);
            values[size-1] = objstrcpy(pool, opt_val);
        }
        select = gui_select_new(list, size, selected);
        free(list);

        ent = (Input**)htab_find_slot(inputs, name, INSERT);
        if (*ent) die("input (name=%s) comes twice\n", name);
        *ent = (Input*)objalloc_alloc(pool, sizeof(Input));
        (*ent)->type = SELECT;
        (*ent)->name = objstrcpy(pool, name);
        (*ent)->widget = select;
        (*ent)->value_size = size;
        (*ent)->values = values;

        return select;
    }
    else if (tidyNodeIsTABLE(node)) {
        int maxx = 0, x = 0, y = 0, size = 0, i;
        TidyNode tr, td;
        widget** widgets = NULL;
        widget* table;

        for (tr = tidyGetChild(node); tr; tr = tidyGetNext(tr)) {
            if (!tidyNodeIsTR(tr)) continue;

            for (td = tidyGetChild(tr); td; td = tidyGetNext(td)) {
                widget* w = widgets_get(pool, tdoc, td);
                size++;
                x++;
                widgets = (widget**)realloc(widgets, sizeof(widget*)*size);
                widgets[size-1] = w;
            }

            if (x > maxx) maxx = x;
            size++;
            y++;
            widgets = (widget**)realloc(widgets, sizeof(widget*)*size);
            widgets[size-1] = NULL;
        }

        table = gui_table_new(maxx, y);
        x = 0;
        y = 0;
        for (i = 0; i < size; i++) {
            widget* w = widgets[i];
            if (w) {
                gui_attach(table, w, x, x+1, y, y+1);
                x++;
            }
            else {
                x = 0;
                y++;
            }
        }

        return table;
    }
    else if (tidyNodeIsIMG(node)) {
        const char* src = (char*)tidyAttrValue(tidyAttrGetSRC(node));
        return gui_img_new(src);
    }
    else {
        die("%s is unknown tag\n", tidyNodeGetName(node));
        return NULL;  /* dummy */
    }
}

widget* hbox_create(widget** widgets, int size) {
    int i;
    widget* hbox;
    if (size == 1) {
        return widgets[0];
    }
    else {
        hbox = gui_hbox_new(size);
        for (i = 0; i < size; i++) {
            gui_pack(hbox, widgets[i]);
        }
        return hbox;
    }
}

widget* widgets_get(objalloc* pool, TidyDoc tdoc, TidyNode form) {
    TidyNode node;
    widget** widgets = NULL;
    widget** vboxes = NULL;
    int vbox_size = 0;
    int size = 0;
    widget* ret;

    for (node = tidyGetChild(form); node; node = tidyGetNext(node)) {
        widget* w;
        if (tidyNodeIsBR(node)) {
            widget* w = hbox_create(widgets, size);
            vbox_size++;
            vboxes = (widget**)realloc(vboxes, sizeof(widget*)*vbox_size);
            vboxes[vbox_size-1] = w;

            free(widgets);
            widgets = NULL;
            size = 0;
        }
        else {
            w = widget_get(pool, tdoc, node);
            size++;
            widgets = (widget**)realloc(widgets, sizeof(widget*)*size);
            widgets[size-1] = w;
        }
    }
    ret = hbox_create(widgets, size);
    free(widgets);

    if (vboxes) {
        widget* vbox = gui_vbox_new(vbox_size+1);
        int i;
        for (i = 0; i < vbox_size; i++) {
            gui_pack(vbox, vboxes[i]);
        }
        gui_pack(vbox, ret);
        free(vboxes);
        return vbox;
    }
    else {
        return ret;
    }
}

static int hash_string_eq (const void *s1_p, const void *s2_p)
{
    const Input* s1 = (const Input*)s1_p;
    const char* s2 = (const char*)s2_p;
    return strcmp(s1->name, s2) == 0;
}

int commit_rc_one(void** ip, void* dummy) {
    Input* input = (Input*)*ip;
    switch (input->type) {
    case TEXT:
        RC_Set(rc, input->name, gui_text_get(input->widget));
        break;
    case CHECKBOX:
        RC_Set(rc, input->name,
               gui_checkbox_get(input->widget) ? input->value : "");
        break;
    case SELECT:
        RC_Set(rc, input->name, input->values[gui_select_get(input->widget)]);
        break;
    case RADIO:
        RC_Set(rc, input->name, input->values[gui_radio_get(input->widget)]);
        break;
    }
    return 1;
}

void commit_rc() {
    htab_traverse(inputs, commit_rc_one, NULL);
}

int free_input(void** ip, void* dummy) {
    Input* input = (Input*)*ip;
    if (input->type == RADIO || input->type == SELECT) {
        free(input->values);
    }
    return 1;
}

void free_inputs() {
    htab_traverse(inputs, free_input, NULL);
}

int main(int argc, char* argv[]) {
    TidyDoc tdoc;
    objalloc* pool;
    widget* window;
    widget* root;
    char* title;
    char* htmlfile;
    char* rcfile;
    TidyNode form;
    char* frame;

    htmlfile = (argc < 2) ? "mwx.html" : argv[1];
    rcfile = (argc < 3) ? "mwx.rc" : argv[2];

    inputs = htab_create(101, htab_hash_string, hash_string_eq, NULL);

    rc = RC_Init();
    RC_Load(rc, rcfile);

    tdoc = tidyCreate();
    tidySetCharEncoding(tdoc, "utf8");
    tidyParseFile(tdoc, htmlfile);
    tidyCleanAndRepair(tdoc);

    pool = objalloc_create();

    title = tidy_get_title(pool, tdoc);

    window = gui_init(&argc, &argv, title);

    form = tidy_search_node(tidyGetBody(tdoc), &tidyNodeIsFORM);
    root = widgets_get(pool, tdoc, form);
    frame = (char*)tidyAttrValue(tidyAttrGetNAME(form));
    if (frame) {
        root = gui_frame_new(root, frame);
    }
    gui_add(window, root);

    gui_start(window);

    RC_Save(rc, rcfile);
    RC_Quit(rc);

    free_inputs();
    objalloc_free(pool);

    return 0;
}
