/**
 * $Id$
 *
 * Copyright (C) shinichiro.h <hamaji@nii.ac.jp>
 *  http://shinh.skr.jp/
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "rc.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

enum { BUFSIZE = 4096 };

static char* mstrcpy(const char* src) {
    char* dst = (char*)malloc(sizeof(char)*(strlen(src)+1));
    strcpy(dst, src);
    return dst;
}

void RCEV_Init(RCEV* rcev) {
    rcev->num = 0;
    rcev->rce = NULL;
}

void RCEV_Add(RCEV* rcev, RCE rce) {
    if ((rcev->num & 15) == 0) {
        rcev->rce = (RCE*)realloc(rcev->rce, sizeof(RCE)*(rcev->num+16));
    }
    rcev->rce[rcev->num] = rce;
    rcev->num++;
}

RCE* RCEV_Get(RCEV* rcev, const char* key) {
    int i;
    for (i = 0; i < rcev->num; i++) {
        RCE* rce = &rcev->rce[i];
        if (rce->arg1 && strcmp(rce->arg1, key) == 0) {
            return rce;
        }
    }
    return NULL;
}

void RCEV_Quit(RCEV* rcev) {
    int i;
    for (i = 0; i < rcev->num; i++) {
        free(rcev->rce[i].arg1);
        free(rcev->rce[i].arg2);
    }
}

RC* RC_Init() {
    RC* rc = (RC*)malloc(sizeof(RC));
    RCEV_Init(&rc->lines);
    RCEV_Init(&rc->adds);
    return rc;
}

void RC_Quit(RC* rc) {
    RCEV_Quit(&rc->lines);
    RCEV_Quit(&rc->adds);
    free(rc);
}

static void tailtrim(char* str) {
    int l;
    for (l = strlen(str)-1; l >= 0; l--) {
        if (isspace(str[l])) str[l] = '\0';
        else return;
    }
}

void RC_Load(RC* rc, const char* file) {
    char buf[BUFSIZE];
    FILE* fp;
    int line_num = 0;
    RCE rce;
    rce.arg1 = NULL;

    fp = fopen(file, "rb");
    if (!fp) {
        fprintf(stderr, "%s not found\n", file);
        return;
    }

    while (fgets(buf, BUFSIZE, fp)) {
        tailtrim(buf);

        if (!rce.arg1 && buf[0] == '\0') continue;

        rce.arg1 = NULL;
        rce.arg2 = NULL;

        if (buf[0] == '#') {
            rce.arg1 = mstrcpy(buf);
        }
        else if (buf[0] != '\0') {
            char* p = strchr(buf, ' ');
            if (!p) {
                rce.arg1 = mstrcpy(buf);
                rce.arg2 = mstrcpy("");
            }
            else {
                *p = '\0';
                p++;
                rce.arg1 = mstrcpy(buf);
                rce.arg2 = mstrcpy(p);
            }
        }

        RCEV_Add(&rc->lines, rce);

        line_num++;
    }
    fclose(fp);
}

void RC_Save(RC* rc, const char* file) {
    int i;
    FILE* fp;

    fp = fopen(file, "wb");

    for (i = 0; i < rc->lines.num; i++) {
        RCE rce = rc->lines.rce[i];
        if (rce.arg2) fprintf(fp, "%s %s\n", rce.arg1, rce.arg2);
        else if (rce.arg1) fprintf(fp, "%s\n", rce.arg1);
        else fputs("\n", fp);
    }

    fputs("\n", fp);

    for (i = 0; i < rc->adds.num; i++) {
        RCE rce = rc->adds.rce[i];
        fprintf(fp, "%s %s\n\n", rce.arg1, rce.arg2);
    }

    fclose(fp);
}

const char* RC_Get(RC* rc, const char* key) {
    RCE* rce;
    rce = RCEV_Get(&rc->lines, key);
    if (!rce) rce = RCEV_Get(&rc->adds, key);
    if (rce) return rce->arg2;
    return NULL;
}

void RC_Set(RC* rc, const char* key, const char* val) {
    RCE* rce;
    rce = RCEV_Get(&rc->lines, key);
    if (!rce) rce = RCEV_Get(&rc->adds, key);
    if (rce) {
        free(rce->arg2);
        rce->arg2 = mstrcpy(val);
    }
    else {
        RCE r;
        r.arg1 = mstrcpy(key);
        r.arg2 = mstrcpy(val);
        RCEV_Add(&rc->adds, r);
    }
}
